<?php

declare(strict_types=1);

$metadata->mapField(
    [
        'type'      => 'integer',
        'name'      => 'hoursWorked',
        'fieldName' => 'hoursWorked',
    ]
);

$metadata->mapField(
    [
        'type'      => 'integer',
        'name'      => 'pricePerHour',
        'fieldName' => 'pricePerHour',
    ]
);
