<?php

declare(strict_types=1);

namespace Doctrine\Tests\Models\DirectoryTree;

/**
 * @Entity
 */
class Directory extends AbstractContentItem
{
    /**
     * @var string
     * @Column(type="string")
     */
    protected $path;

    public function setPath(string $path): void
    {
        $this->path = $path;
    }

    public function getPath(): string
    {
        return $this->path;
    }
}
