<?php

declare(strict_types=1);

$metadata->mapField(
    [
        'id'         => true,
        'fieldName'  => 'id',
        'type'       => 'integer',
        'columnName' => 'id',
    ]
);

//$metadata->setIdGeneratorType(ClassMetadataInfo::GENERATOR_TYPE_AUTO);
