<?php

declare(strict_types=1);

namespace Doctrine\Tests\ORM\Hydration;

use Doctrine\ORM\EntityManagerInterface;
use Doctrine\Tests\OrmTestCase;

class HydrationTestCase extends OrmTestCase
{
    /** @var EntityManagerInterface */
    protected $entityManager;

    protected function setUp(): void
    {
        parent::setUp();
        $this->entityManager = $this->getTestEntityManager();
    }
}
