<?php

/*
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * This software consists of voluntary contributions made by many individuals
 * and is licensed under the MIT license. For more information, see
 * <http://www.doctrine-project.org>.
 */

require_once __DIR__ . '/../Annotation.php';
require_once __DIR__ . '/../Entity.php';
require_once __DIR__ . '/../Embeddable.php';
require_once __DIR__ . '/../Embedded.php';
require_once __DIR__ . '/../MappedSuperclass.php';
require_once __DIR__ . '/../InheritanceType.php';
require_once __DIR__ . '/../DiscriminatorColumn.php';
require_once __DIR__ . '/../DiscriminatorMap.php';
require_once __DIR__ . '/../Id.php';
require_once __DIR__ . '/../GeneratedValue.php';
require_once __DIR__ . '/../Version.php';
require_once __DIR__ . '/../JoinColumn.php';
require_once __DIR__ . '/../JoinColumns.php';
require_once __DIR__ . '/../Column.php';
require_once __DIR__ . '/../OneToOne.php';
require_once __DIR__ . '/../OneToMany.php';
require_once __DIR__ . '/../ManyToOne.php';
require_once __DIR__ . '/../ManyToMany.php';
require_once __DIR__ . '/../Table.php';
require_once __DIR__ . '/../UniqueConstraint.php';
require_once __DIR__ . '/../Index.php';
require_once __DIR__ . '/../JoinTable.php';
require_once __DIR__ . '/../SequenceGenerator.php';
require_once __DIR__ . '/../CustomIdGenerator.php';
require_once __DIR__ . '/../ChangeTrackingPolicy.php';
require_once __DIR__ . '/../OrderBy.php';
require_once __DIR__ . '/../NamedQueries.php';
require_once __DIR__ . '/../NamedQuery.php';
require_once __DIR__ . '/../HasLifecycleCallbacks.php';
require_once __DIR__ . '/../PrePersist.php';
require_once __DIR__ . '/../PostPersist.php';
require_once __DIR__ . '/../PreUpdate.php';
require_once __DIR__ . '/../PostUpdate.php';
require_once __DIR__ . '/../PreRemove.php';
require_once __DIR__ . '/../PostRemove.php';
require_once __DIR__ . '/../PostLoad.php';
require_once __DIR__ . '/../PreFlush.php';
require_once __DIR__ . '/../FieldResult.php';
require_once __DIR__ . '/../ColumnResult.php';
require_once __DIR__ . '/../EntityResult.php';
require_once __DIR__ . '/../NamedNativeQuery.php';
require_once __DIR__ . '/../NamedNativeQueries.php';
require_once __DIR__ . '/../SqlResultSetMapping.php';
require_once __DIR__ . '/../SqlResultSetMappings.php';
require_once __DIR__ . '/../AssociationOverride.php';
require_once __DIR__ . '/../AssociationOverrides.php';
require_once __DIR__ . '/../AttributeOverride.php';
require_once __DIR__ . '/../AttributeOverrides.php';
require_once __DIR__ . '/../EntityListeners.php';
require_once __DIR__ . '/../Cache.php';
